import java.lang.Math;

//NB : compilation pour javadoc:    javadoc -d Documentation Integrateur.java

/**
 * Classe contenant un ensemble de methodes pour integrer un probleme donne par l'equation de mouvement "eqMvmt()" (qui est implementee dans une classe qui implemente UsineEqMvmt.java). On utilise une methode runge-kutta d'ordre 4 "rk4()". Mais aussi runge-kutta a pas variable "rk4pv()".
 *
 *@author Nicolas DELSATE
 *@version
 *
 * Derniere modification :
 *  03/01/2006
 */
public class Integrateur{

 /**
   *@param h          pas d'integration courant initialise a 0 dans le constructeur
   *@param taille     taille du tableau que l'on passe en parametre du constructeur, equivalent au nombre d'equation du mouvement
   *@param parametres tableau de parametres passe a la methode "eqMvmt()"
   *@param nbrParam   taille du tableau parametres          
   *@param xNew       tableau servant a l'adaptation du pas, utilise dans rk4pv
   *@param tNew       temps au dernier pointt trouve par l'algo, de maniere generale = tNew+h ; au debut il est = 0
   *@param tFinal     temps auquel on demande une reponse, il sera de maniere generale = tFinal+intTps ; au debut il est = 0
   *@param erreur     valeur de la derniere erreur commise : sert pour l'adaptation du pas
   *@param mvmt       contient le "nom" de l'equation du mouvement; on fera acces a l'equation en faisant mvmt.eqMvmt
   */
   
  double h;//le pas utilise dans l'integration courante, initialement =0


  int taille; //taille du tableau que l'on passe en parametre du constructeur, equivalent au nombre d'equation du mouvement


  double[] parametres; //tableau de parametres
  int nbrParam; //nombre de parametres


  //Le tableau et le tps suivant servent a l'adaptation du pas
  //variables utilisees dans rk4pv !!!
  double[] xNew;

  //tps ou on calcule le tableau ci-dessus
  double tNew; //tps au dernier pt trouve par l'algo, de maniere generale = tNew+h ; au debut il est = 0

  //tps auquel on demande 1reponse, il sera de maniere generale = tFinal+intTps ; au debut il est = 0
  double tFinal;

  //valeur de la derniere erreur commise : sert pour l'adaptation du pas
  double erreur;

  InterfaceEqMvmt mvmt;//contient le "nom" de l'equation du mouvement; on fera acces a l'equation en faisant mvmt.eqMvmt

 


  // constructeur de la classe
  /**
   *@param nbrVariable nombre d'equation du mouvement (ou de maniere nombre de variables)
   *@param tabParametres tableau contenant les differents parametres
   *@param fEqMvmt nom du fichier contenant la methode qui implemente la methode eqMvmt de l'interface InterfaceEqMvmt
   */
  public Integrateur(int nbrVariable,double[] tabParametres,String fEqMvmt) {
    taille=nbrVariable;//taille du tableau

    nbrParam=tabParametres.length;//taille du tableau de parametre
    parametres=new double[nbrParam]; //allocation de memoire pr tableau de parametres
    for(int i=0;i<nbrParam;i++) {
      parametres[i]=tabParametres[i];//tableau de parametre
    }

    //initialisation d'un pas nul
    h=0;

    //creation de mvmt de type InterfaceEqMvmt
    mvmt=UsineEqMvmt.nouveau(fEqMvmt);

    //initialisation des variables utilisees dans rk4pv !!!
    xNew=new double[taille];
    tNew=0;
    tFinal=0;
  }






  //////////////////////////////////////////////////////////////////
  //                          rk4                                 //
  //////////////////////////////////////////////////////////////////

  /**  
   * methode qui utilise une methode runge-kutta a pas fixe d'ordre 4
   * pour calculer a partir d'un vecteur x et d'un pas une solution xPlush
   * Formule generale : x(t+h)=x(t)+h/6*(k1+2*k2+2*k3+k4);
   * @param  x  un tableau de la forme [X1,X2,...]
   * @param  hh pas d'integration
   * @return    un tableau de la forme [X1,X2,...]
   */

  public double[] rk4(double[] x,double hh){
    //formule generale : x(t+h)=x(t)+h/6*(k1+2*k2+2*k3+k4);

    //les 4 facteurs utilises dans la methode runge-kutta
    double[] k1=new double[taille];
    double[] k2=new double[taille];
    double[] k3=new double[taille];
    double[] k4=new double[taille];

    //calcul du facteur k1
    k1=mvmt.eqMvmt(x,parametres);

    //tabeau transit utilise pour calculer les autres facteurs
    double[] transit=new double[taille];

    //calcul du facteur k2
    for(int i=0; i<taille; i++){
      transit[i]=x[i]+hh/2*k1[i];
    }
    k2=mvmt.eqMvmt(transit,parametres);


    //calcul du facteur k3
    for(int i=0; i<taille; i++){
      transit[i]=x[i]+hh/2*k2[i];
    }
    k3=mvmt.eqMvmt(transit,parametres);

    //calcul du facteur k4
    for(int i=0; i<taille; i++){
      transit[i]=x[i]+hh*k3[i];
    }
    k4=mvmt.eqMvmt(transit,parametres);

    
    //la reponse x(t+hh)
    double[] xPlush=new double[taille];
    for(int i=0;i<taille;i++){
      xPlush[i]=x[i]+hh/6*(k1[i]+2*k2[i]+2*k3[i]+k4[i]);
    }

    return xPlush;
  }






  //////////////////////////////////////////////////////////////////
  //                          rk4pv                               //
  //////////////////////////////////////////////////////////////////

  /**
   * runge-kutta d'ordre 4 a pas variable avec une precision a preciser
   * Cette methode adapte le pas lors de son integration mais aussi adapte le pas pour que la derniere etape donne exactement le resultat au temps "intTps" voulu sans faire d'interpolation !
   * @param  x           un tableau de la forme [X1,X2,...]
   * @param  intTps      interval de temps durant lequel on fait l'integration, c'est apres ce temps qu'on veut une reponse. 
   * @param  precision   precision sur la difference relative entre deux resultats. Si l'on atteind cette precision, on renvoit la solution. Sinon, on diminue le pas et on recommence
   * @param  vieuxCalcul =FALSE si on veut faire un nouveau calcul et =TRUE si on continue un ancien calcul; dans ce cas, on ne tient pas compte du "x" passe en argument
   * @return             un tableau de la forme [X1,X2,...]
   */

  public double[] rk4pv(double[] x,double intTps,double precision, boolean vieuxCalcul){

    //test preliminaire
    if(intTps<=0) {
      System.out.println("!?!  intTps doit etre >0   !?!");
    }


    //creation de variables propre a la methode
    //x(t+h)
    double[] x1Plush=new double[taille];

    //x(t+h/2)
    double[] x2aPlushDemi=new double[taille];

    //x(t+h) obtenu a partir de x(t+h/2)
    double[] x2Plush=new double[taille];

    //erreur relative commise lors de l'integration
    double[] delta = new double[taille]; //contiendra |x1Plush-x2Plush|/(|x1Plush|+|x2Plush|+x') i.e l'erreur relative commise lors de l'integration 


    //initialisation du pas maximum h
    double hmax=intTps;


    //mise a jour de la variable GLOBALE tFinal
    //tps a l'interpolation ou l'on demande la solution
    tFinal=tFinal+intTps;



    //si le pas h est nul, faut bien lui donner qq chose
    //En fait, lorsqu'on appel la fonction la premiere fois, le pas h est 
    //egale a 0. Il faut donc l'initialiser a qq chose.
    //On donne l'interval de temps d'integration (hmax), donc, on initialise 
    //ce pas h avec ce intTps.


    if(h==0){
      h=hmax;
    }

    //sinon, on ne tient pas compte de intTps pour determiner h, on utilise le 
    //pas h qui est declare en global car on a deja appele la methode 
    //precedemment.


    if(! vieuxCalcul){ //c'est un nouveau calcul

      //initialisation de xNew[],variable utilisee ds ts les calculs suivants
      for(int i=0;i<taille;i++){
        xNew[i]=x[i];
      }

      //initialisation de l'erreur pour pouvoir, lors du premier passage 
      //dans la boucle do{...}while(erreur>precision), commencer avec un pas 
      //qui vaut h (en effet, cette boucle commence par une adaptation du 
      //pas !
      erreur=precision*Math.pow(0.9,5);
    }
    //else on est dans la suite d'un calcul deja commence donc, ne tient pas 
    //compte de x donne en parametre et on redemarre avec l'ancien xNew


    //pour pouvoir rentrer dans la boucle qui suit
    boolean dernier=false;
  
    while(! dernier) { //"presque" equivalent a tNew<tFinal)

      //recherche du pas optimal pour calculer le xNew
      do { 
        dernier=false;
 
        //adaptation du pas : - augmente si on entre dans la boucle car 
        //                      erreur<precision. A cause du facteur 
        //                      (de securite) O.9, il y a une augmentation 
        //                      lorsque erreur<1.6935088
        //                    - diminue si on viens de la fin de la boucle car 
        //                      erreur>precision
        h=h*0.9*Math.pow(precision/erreur,0.2);


        //si le pas calcule est plus grand que le pas maximum autorise
        if(h>hmax) {
          h=hmax;
        //---Debugging----
        //System.out.println("-----------h>hmax => h=hmax=" + h);
        //---Debugging----
      }
  
      //si le pas h calcule est tel qu'avec celui-ci, on donnera une 
      //solution a un temps plus grand que celui demande, alors, on adapte le
      //pas en l'egalant au pas qu'il faudrait en partant de tNew pour 
      //pouvoir finir au temps exacte demande (tFinal).
      if(h> tFinal-tNew ) {
        h=tFinal-tNew;

        dernier=true;//puisqu'on adapte le pas de cette facon, c'est 
                     //qu'apres, il n'y aura plus d'autre pas.
      }

      //calcul de x(t+h)
      x1Plush=rk4(xNew,h);

      //calcul de x(t+h/2)
      x2aPlushDemi=rk4(xNew,h/2);

      //calcul de x(t+h/2+h/2)
      x2Plush=rk4(x2aPlushDemi,h/2);


      //pour que l'erreur ne devienne pas infinie lors du passage d'un cote
      // a l'autre de l'axe des x, on ajoutera yPrime*h au denominateur
      double[] yPrime = new double[taille];
      yPrime = mvmt.eqMvmt(x2Plush,parametres);


      //calcul de l'erreur commise
      for(int i=0;i<taille;i++){
        delta[i]=Math.abs( x2Plush[i]-x1Plush[i] ) / ( Math.abs(x2Plush[i]) + Math.abs(x1Plush[i]) + Math.abs(yPrime[i]*h));
      }

      erreur=Math.abs(maximum(delta));



    } while(erreur>precision);

    //arriver ici, on a un pas optimal h et on met enfin a jour notre xNew et
    //tNew

    //actualiastion de xNew
    for(int i=0;i<taille;i++){
      xNew[i]=x2Plush[i] + (x2Plush[i]-x1Plush[i])/15;
      //xNew a alors une erreur d'ordre 6 au lieu de l'ordre 5
    }

    //acutalisation de tNew
    tNew=tNew+h;
  

    }//fin de while(tNew<tFinal)


    double[] xResultatT=new double[taille];

    for(int i=0; i<taille ; i++) {
      xResultatT[i]=xNew[i];
    }

    return xResultatT;
  }



  /**
   * runge-kutta d'ordre 4 a pas variable avec une precision predefinie=1.0e-3
   * @param  x      un tableau de la forme [Q1,Q2,P1,P2]
   * @param  intTps interval de temps durant lequel on fait l'integration, c'est apres ce temps qu'on veu une reponse
   * @return        un tableau de la forme [Q1,Q2,P1,P2]
   */

  public double[] rk4pv(double[] x,double intTps){
    double precision=1.0e-6;
    double[] xResultatT=new double[taille];
    xResultatT=rk4pv(x,intTps,precision);
    return xResultatT;
  }



  /**
   * appel la fonction rk4pv(double[] x, double deltaT, double precision, boolean vieuxCalcul) avec precision fixee a 1.0e-6
   * Voir rk4pv(double[] x,double deltaT, double precision, boolean vieuxCalcul)
   */

  public double[] rk4pv(double[] x,double deltaT, boolean vieuxCalcul){
    double[] xResultatT=new double[taille]; //tableau de resultat au temps deltaT exactement !!!
    double precision=1.0e-6;   
    xResultatT=rk4pv(x,deltaT,precision,vieuxCalcul);
    return xResultatT;
  }



  /**
   * appel la fonction rk4pv(double deltaT, double precision) avec une precision fixee a 1.0e-6
   * Voir rk4pv(double deltaT, double precision)
   */

  public double[] rk4pv(double deltaT){
    double precision=1.0e-6;
    double[] xResultatT=new double[taille]; //tableau de resultat au temps deltaT exactement !!!
    xResultatT=rk4pv(deltaT,precision);
    return xResultatT;
  }



  /**
   * appel la fonction rk4pv(double[] x, double deltaT, double precision, boolean vieuxCalcul) sachant que on est dans la suite d'un vieux calcul. Donc, on passe un x bidon comme premier parametre et "true" pour le parametre boolean
   * Voir rk4pv(double[] x,double deltaT, double precision, boolean vieuxCalcul)
   */

  public double[] rk4pv(double deltaT, double precision){
    double[] xResultatT=new double[taille]; //tableau de resultat au temps deltaT exactement !!!
    double[] xBidon=new double[taille];
    for(int i=0;i<taille;i++){
      xBidon[i]=0;
    }
    xResultatT=rk4pv(xBidon,deltaT,precision,true);
    return xResultatT;
  }



  /**
   * appel la fonction rk4pv(double[] x, double deltaT, double precision, boolean vieuxCalcul) avec "false" pour le parametre boolean
   * Voir rk4pv(double[] x,double deltaT, double precision, boolean vieuxCalcul)
   */

  public double[] rk4pv(double[] x,double deltaT, double precision){
    double[] xResultatT=new double[taille]; //tableau de resultat au temps deltaT exactement !!!
    xResultatT=rk4pv(x,deltaT,precision,false);
    return xResultatT;
  }






  //////////////////////////////////////////////////////////////////
  //                       interpolation                          //
  //////////////////////////////////////////////////////////////////

  /**
   * Interpolation par une droite
   * @param y1        la valeur de la fonction evalue en t1
   * @param y2        la valeur de la fonction evalue en t2
   * @param tInterpol sachant que t est le temps auquel on veut l'interpolation, tInterpol=t-t1
   * @param tTotal    t2-t1
   * @return   l'interpolation au temps t
   */

  public double[] interpolation(double[] y1, double[] y2, double tInterpol, double tTotal){

    double[] resultatInterpolation=new double[taille];

    for(int i=0;i<taille;i++){

     resultatInterpolation[i]=y1[i]+(y2[i]-y1[i])*(tInterpol)/(tTotal);
    }

    return resultatInterpolation;
  }






  //////////////////////////////////////////////////////////////////
  //                       maximum                                //
  //////////////////////////////////////////////////////////////////

  /**
   * Calcul de la valeur de l'element le plus grand en valeur absolue d'un tableau
   * @param tab      un tableau de double
   * @return         la valeur de l'element le plus grand en valeur absolue d'un tableau
   */

  public double maximum(double[] tab){
    double maxTab=Math.abs(tab[0]);
    int longueur=tab.length;

    for(int i=1;i<longueur;i++){
      if(maxTab<Math.abs(tab[i])){
         maxTab=Math.abs(tab[i]);
      }
    }
    return maxTab;
  }


}//fin de public class
